package es.uca.xmleye.gui;

/*
 Copyright (C) 2005-2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import javax.swing.*;
import javax.swing.Timer;
import javax.swing.tree.TreePath;
import javax.swing.border.EmptyBorder;
import javax.swing.event.*;
import java.util.concurrent.Callable;

import net.infonode.tabbedpanel.TabDragEvent;
import net.infonode.tabbedpanel.TabEvent;
import net.infonode.tabbedpanel.TabListener;
import net.infonode.tabbedpanel.TabRemovedEvent;
import net.infonode.tabbedpanel.TabStateChangedEvent;

import org.w3c.dom.*;

import es.uca.xmleye.ParsingException;
import es.uca.xmleye.SearchRequest;
import es.uca.xmleye.events.FileChangeEvent;
import es.uca.xmleye.events.FileChangeListener;
import es.uca.xmleye.formats.FormatDescriptorException;
import es.uca.xmleye.gui.events.SearchListener;
import es.uca.xmleye.gui.util.ResourceBackedAction;
import es.uca.xmleye.gui.util.ResourceBackedButton;
import es.uca.xmleye.gui.util.ExtensionFileFilter;
import es.uca.xmleye.gui.util.RadioSelectMenu;
import es.uca.xmleye.gui.util.ResourceBackedMenu;
import es.uca.xmleye.presentation.DocumentPresentationModel;
import es.uca.xmleye.presentation.WndMainPresentationModel;
import es.uca.xmleye.presentation.WndFormatsPresentationModel;
import es.uca.xmleye.util.UnaryFunctor;
import es.uca.xmleye.util.Pair;
import es.uca.xmleye.util.MultipleCallbackExecutorService;
import es.uca.xmleye.xslt.StylesheetRepositoryException;
import es.uca.xmleye.xslt.TransformationException;

public class WndMain extends JFrame implements TreeSelectionListener,
		SearchListener, FileChangeListener, HyperlinkListener, TabListener {
	/**
	 * 
	 */
	private static final long serialVersionUID = 3584670885171884520L;

	// Intervalo de sincronización de estado
	private final static int INTERVALO_SINC_ESTADO = 250;

	// Aplicación principal
	private final Application aplicacion;

	// Modelo de presentación
	private final WndMainPresentationModel modPresentacion;
	// Modelo de presentación de formatos
	private final WndFormatsPresentationModel modPresentacionFormatos;

	// Servicio de ejecución de tareas
	private final MultipleCallbackExecutorService servEjecucion;

	// Recursos para internacionalización
	private static final ResourceBundle res;
	private static final ResourceBundle errRes;
	private static final ResourceBundle miscRes;

	// Acciones
	private ResourceBackedAction accionActivarAutoImport;
	private ResourceBackedAction accionAcercaDe;
	private ResourceBackedAction accionPrimerHijo;
	private ResourceBackedAction accionContraerTodo;
	private ResourceBackedAction accionEditar;
	private ResourceBackedAction accionSalir;
	private ResourceBackedAction accionExpandirTodo;
	private ResourceBackedAction accionBuscar;
	private ResourceBackedAction accionBuscarSig;
	private ResourceBackedAction accionHermanoSig;
	private ResourceBackedAction accionAbrir;
	private ResourceBackedAction accionPadre;
	private ResourceBackedAction accionHermanoAnt;
	private ResourceBackedAction accionSelHojaPreproc;
	private ResourceBackedAction accionSelHojaVista;
	private ResourceBackedAction accionEditarFormatos;

	// / Tamaño mínimo
	final static int ANCHO_MIN = 640;
	final static int ALTO_MIN = 480;

	// Subventanas
	private WndFind vntBusqueda;

	// Elementos a sincronizar de vez en cuando
	private JMenu menuFichero;
	private JMenu menuPreferencias;
	private RadioSelectMenu menuHojaPreprocesado;
	private RadioSelectMenu menuHojaVisualizacion;
	private JCheckBoxMenuItem chkHabilitarAutoImportar;
	private JLabel estadoPrograma;
	private JProgressBar progresoAccion;
	private DocumentList listaDocs;

	static {
		res = ResourceBundle.getBundle("es.uca.xmleye.gui.l10n.ResWndMain",
				Locale.getDefault());
		errRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.ErrorMessages",
				Locale.getDefault());
		miscRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.MiscMessages",
				Locale.getDefault());
	}

	/**
	 * Crea la ventana principal de la aplicación y la muestra.
	 * 
	 * @param titulo
	 *            Título de la ventana.
	 * @param aplicacion
	 *            Moderador de la aplicación.
	 * @param docsPorAbrir
	 *            Documentos a abrir (pasados por línea de órdenes).
	 * @throws FormatDescriptorException
	 * @throws StylesheetRepositoryException
	 * @throws TransformationException
	 * @throws ParsingException
	 * @throws StylesheetRepositoryException
	 */
	public WndMain(Application aplicacion, List<File> fichPorAbrir,
			WndMainPresentationModel modPresentacion,
			WndFormatsPresentationModel modPresentacionFormatos)
			throws FormatDescriptorException, ParsingException,
			TransformationException, StylesheetRepositoryException {
		super(modPresentacion.getTitulo(null));

		this.aplicacion = aplicacion;
		this.servEjecucion = new MultipleCallbackExecutorService();
		this.modPresentacionFormatos = modPresentacionFormatos;
		this.modPresentacion = modPresentacion;

		inicPropiedades();
		inicComponentes();
		inicAcciones();
		inicBarraMenu();
		inicBarraHerramientas();
		inicBarraEstado();

		pack();
		sincronizarDocumento();
		sincronizarEstado();

		setVisible(true);

		// Hacemos que abra todos los documentos pasados por la línea de
		// órdenes
		for (File fich : fichPorAbrir) {
			abrirFichero(fich, null);
		}

	}

	public void valueChanged(TreeSelectionEvent e) {
		final DocumentPresentationModel docActual = getDocumentoActual();
		final TreePath nuevaSel = e.getNewLeadSelectionPath();

		servEjecucion.ejecutar(new Callable<String>() {
			public String call() throws Exception {
				return docActual.solicitudCambioSeleccion(nuevaSel);
			}
		}, new UnaryFunctor<String>() {
			public void ejecutar(String nuevoXHTML) {
				docActual.finCambioSeleccion(nuevoXHTML, nuevaSel);
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						sincronizarDocumento();
					}
				});
			}
		}, new UnaryFunctor<Exception>() {
			public void ejecutar(Exception ex) {
				ex.printStackTrace();
				MessageManagerSingleton.getInstancia().mostrarError(
						ex.getMessage());
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						sincronizarDocumento();
					}
				});
			}
		});
	}

	public void busquedaLanzada(final SearchRequest ev) {
		final DocumentPresentationModel docActual = getDocumentoActual();

		servEjecucion.ejecutar(new Callable<NodeList>() {
			public NodeList call() throws Exception {
				return docActual.solicitudBusqueda(ev);
			}
		}, new UnaryFunctor<NodeList>() {
			public void ejecutar(final NodeList res) {
				docActual.finBusqueda(res, ev);
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						if (res.getLength() == 0) {
							MessageManagerSingleton
									.getInstancia()
									.mostrarError(
											miscRes.getString("find_nomatches"));
						}
						sincronizarDocumento();
					}
				});
			}
		}, new UnaryFunctor<Exception>() {
			public void ejecutar(Exception ex) {
				ex.printStackTrace();
				MessageManagerSingleton.getInstancia().mostrarError(
						ex.getMessage());
			}
		});
	}

	public void ficheroCambiado(FileChangeEvent e) {
		abrirFichero(e.getFicheroCambiado(), null);
	}

	public void hyperlinkUpdate(HyperlinkEvent e) {
		if (e.getEventType() != HyperlinkEvent.EventType.ACTIVATED)
			return;

		final String link = e.getDescription();
		final Matcher m = Pattern.compile("([^#]+)(?:#xpointer\\((.+)\\))?")
				.matcher(link);

		if (!m.matches())
			return;

		final String ruta = m.group(1);
		final String xpointer = m.group(2);
		File f = new File(ruta);

		if (f.canRead()) {
			// Ruta relativa respecto a actual / absoluta:
			// no hay que hacer nada más
			abrirFichero(f, xpointer);
		} else {

			// No sirve la dirección tal y como está: usaremos el directorio
			// base del documento actual.
			DocumentPresentationModel docActual = getDocumentoActual();
			final String dirBase = docActual.getRutaBase();
			f = new File(dirBase + File.separator + ruta);

			abrirFichero(f, xpointer);
		}
	}

	/* --- METODOS PROTEGIDOS -------------------------------- */

	/**
	 * Inicializa las propiedades del diálogo.
	 */
	protected void inicPropiedades() {
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		JPanel panel = new JPanel();
		panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
		setContentPane(panel);

		/** Avisaremos al controlador de cuando el usuario cierre esta ventana */
		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent ev) {
				aplicacion.close();
			}
		});

		/** Aseguramos un tamaño mínimo de la ventana */

		// Esto sólo funciona a partir de Mustang (J2SE 6) como debería ser
		setPreferredSize(new Dimension(ANCHO_MIN, ALTO_MIN));

		/*
		 * Usamos esta solución a un bug que ha permanecido desde J2SE 2
		 * (publicado en 09/03/2000, resuelto en Mustang (!) ). No es ideal,
		 * pero es la mejor que hay.
		 * 
		 * (Ver http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4320050)
		 */
		addComponentListener(new ComponentAdapter() {
			@Override
			public void componentResized(ComponentEvent e) {
				int w = getWidth();
				int h = getHeight();
				if (w < ANCHO_MIN || h < ALTO_MIN) {
					pack();
				}
			}
		});
	}

	protected void inicComponentes() throws ParsingException,
			TransformationException {
		// LISTA DE DOCUMENTOS
		listaDocs = new DocumentList(modPresentacion);
		listaDocs.setPreferredSize(new Dimension(ANCHO_MIN, 350));
		listaDocs.addTabListener(this);
		getContentPane().add(listaDocs);
	}

	protected void inicAcciones() {
		/* Las acciones simples se crean aquí como clases anónimas. */
		accionAcercaDe = new ResourceBackedAction(res, "action_about") {
			/**
			 * 
			 */
			private static final long serialVersionUID = 1791195432791531091L;

			public void actionPerformed(ActionEvent e) {
				new WndAbout(WndMain.this);
			}
		};
		accionPrimerHijo = new ResourceBackedAction(res, "action_child", res
				.getString("action_child_i")) {
			/**
			 * 
			 */
			private static final long serialVersionUID = -9011637049355624750L;

			public void actionPerformed(ActionEvent e) {
				listaDocs.getVistaTabActual().irPrimerHijo();
			}
		};
		accionContraerTodo = new ResourceBackedAction(res, "action_collapse") {
			/**
			 * 
			 */
			private static final long serialVersionUID = -3673103902091478796L;

			public void actionPerformed(ActionEvent e) {
				listaDocs.getVistaTabActual().contraerTodo();
			}
		};
		accionEditar = new ResourceBackedAction(res, "action_edit", res
				.getString("action_edit_i")) {
			/**
			 * 
			 */
			private static final long serialVersionUID = -8958694305808734851L;

			public void actionPerformed(ActionEvent e) {
				try {
					getDocumentoActual().editarDocumento();
				} catch (Exception ex) {
					ex.printStackTrace();
					MessageManagerSingleton.getInstancia().mostrarError(
							ex.getMessage());
				}
			}
		};
		accionExpandirTodo = new ResourceBackedAction(res, "action_expand") {
			/**
			 * 
			 */
			private static final long serialVersionUID = 1461654585040886733L;

			public void actionPerformed(ActionEvent e) {
				listaDocs.getVistaTabActual().expandirTodo();
			}
		};
		accionHermanoSig = new ResourceBackedAction(res, "action_nextsibling",
				res.getString("action_nextsibling_i")) {
			/**
			 * 
			 */
			private static final long serialVersionUID = 5855016252380449128L;

			public void actionPerformed(ActionEvent e) {
				listaDocs.getVistaTabActual().irSigHermano();
			}
		};
		accionPadre = new ResourceBackedAction(res, "action_parent", res
				.getString("action_parent_i")) {
			/**
			 * 
			 */
			private static final long serialVersionUID = 8506197212134279952L;

			public void actionPerformed(ActionEvent e) {
				listaDocs.getVistaTabActual().irPadre();
			}
		};
		accionHermanoAnt = new ResourceBackedAction(res, "action_prevsibling",
				res.getString("action_prevsibling_i")) {
			/**
			 * 
			 */
			private static final long serialVersionUID = -4688738546929618943L;

			public void actionPerformed(ActionEvent e) {
				listaDocs.getVistaTabActual().irAntHermano();
			}
		};

		/**
		 * Acción de (des)activación de importación automática.
		 */
		accionActivarAutoImport = new ResourceBackedAction(res,
				"action_toggleautoimport") {
			/**
			 * 
			 */
			private static final long serialVersionUID = 3595025733061926016L;

			public void actionPerformed(ActionEvent e) {
				modPresentacion.cambioAutoImportacion();
				sincronizarDocumento();
			}
		};

		accionSalir = new ExitAction();
		accionBuscar = new FindAction();
		accionBuscarSig = new FindNextAction();
		accionAbrir = new OpenAction();

		accionSelHojaPreproc = new ChangePreprocessingStylesheetAction();
		accionSelHojaVista = new ChangeViewStylesheetAction();

		accionEditarFormatos = new EditFormatsAction();
	}

	/**
	 * Inicializa la barra de menús.
	 */
	protected void inicBarraMenu() {
		JMenuBar menu = new JMenuBar();
		menu.setPreferredSize(new Dimension(350, 20));

		// INICIO MENÚ FICHERO
		{
			menuFichero = new ResourceBackedMenu(res, "menu_file");
			sincronizarMenuFichero(menuFichero);
			menu.add(menuFichero);
		}

		// INICIO MENÚ NAVEGAR
		{
			JMenu menuNavegar = new ResourceBackedMenu(res, "menu_browse");
			menuNavegar.add(new JMenuItem(accionBuscar));
			menuNavegar.add(new JMenuItem(accionBuscarSig));
			menuNavegar.addSeparator();
			menuNavegar.add(new JMenuItem(accionPadre));
			menuNavegar.add(new JMenuItem(accionPrimerHijo));
			menuNavegar.add(new JMenuItem(accionHermanoAnt));
			menuNavegar.add(new JMenuItem(accionHermanoSig));
			menuNavegar.addSeparator();
			menuNavegar.add(new JMenuItem(accionExpandirTodo));
			menuNavegar.add(new JMenuItem(accionContraerTodo));
			menu.add(menuNavegar);
		}

		// INICIO MENÚ PREFERENCIAS
		{
			menuPreferencias = new ResourceBackedMenu(res, "menu_preferences");
			sincronizarMenuPreferencias(menuPreferencias);
			menu.add(menuPreferencias);
		}

		// INICIO MENÚ AYUDA
		{
			JMenu menuHelp = new ResourceBackedMenu(res, "menu_help");
			menuHelp.add(new JMenuItem(accionAcercaDe));
			menu.add(Box.createHorizontalGlue());
			menu.add(menuHelp);
		}

		setJMenuBar(menu);
	}

	/**
	 * Inicializa la barra de herramientas.
	 */
	protected void inicBarraHerramientas() {
		JToolBar barraHerramientas = new JToolBar(res
				.getString("toolbar_main_title"));
		barraHerramientas.setBorder(new EmptyBorder(1, 1, 1, 1));

		barraHerramientas.add(new ResourceBackedButton(res, accionAbrir,
				"action_open"));
		barraHerramientas.add(new ResourceBackedButton(res, accionEditar,
				"action_edit"));
		barraHerramientas.addSeparator();
		barraHerramientas.add(new ResourceBackedButton(res, accionBuscar,
				"action_find"));
		barraHerramientas.add(new ResourceBackedButton(res, accionPadre,
				"action_parent"));
		barraHerramientas.add(new ResourceBackedButton(res, accionPrimerHijo,
				"action_child"));
		barraHerramientas.add(new ResourceBackedButton(res, accionHermanoAnt,
				"action_prevsibling"));
		barraHerramientas.add(new ResourceBackedButton(res, accionHermanoSig,
				"action_nextsibling"));
		barraHerramientas.addSeparator();
		barraHerramientas.add(new ResourceBackedButton(res, accionSalir,
				"action_exit"));
		barraHerramientas.add(Box.createHorizontalGlue());

		getContentPane().add(barraHerramientas, 0);
	}

	/**
	 * Inicializa la barra de estado.
	 */
	protected void inicBarraEstado() {
		final JPanel panelBarraEstado = new JPanel(new BorderLayout());
		panelBarraEstado.setBorder(BorderFactory.createEmptyBorder(3, 6, 3, 3));
		panelBarraEstado.setMaximumSize(new Dimension(1500, 40));

		// Mensaje de estado
		estadoPrograma = new JLabel();
		estadoPrograma.setPreferredSize(new Dimension(300, 20));
		panelBarraEstado.add(estadoPrograma, BorderLayout.WEST);

		// Progreso de acción
		progresoAccion = new JProgressBar(SwingConstants.HORIZONTAL, 0, 100);
		progresoAccion.setIndeterminate(false);
		progresoAccion.setValue(100);
		panelBarraEstado.add(progresoAccion, BorderLayout.EAST);

		// Comprobación en cada medio segundo
		new Timer(INTERVALO_SINC_ESTADO, new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				sincronizarEstado();
			}
		}).start();

		getContentPane().add(panelBarraEstado);
	}

	/* --------------- METODOS PRIVADOS ------------------- */

	/* -- SINCRONIZACION: NIVEL DE APLICACION -- */

	/**
	 * (Re)construye el menú archivo, suponiendo que las acciones ya están
	 * creadas.
	 * 
	 * @param menuFichero
	 *            Menú al que modificar sus elementos.
	 */
	private void sincronizarMenuFichero(JMenu menuFichero) {
		menuFichero.removeAll();

		menuFichero.add(new JMenuItem(accionAbrir));
		menuFichero.add(new JMenuItem(accionEditar));

		// Documentos recientes
		List<String> hist = modPresentacion.getEntradasHistorial();
		// Anything longer than this many characters should be shortened
		final int LONG_LABEL = 20;
		
		if (hist.size() > 0) {
			menuFichero.addSeparator();

			for (ListIterator<String> it = hist.listIterator(); it.hasNext();) {
				final String elem = it.next();
				
				String label;
				if (elem.length() < LONG_LABEL) {
					label = elem;
				} else {
					label = "... "
						+ elem.substring(Math.max(0, elem.length() - LONG_LABEL));
				}

				JMenuItem elemHistorial = new JMenuItem(label, elem.length());
				elemHistorial.addActionListener(new ActionListener() {
					public void actionPerformed(ActionEvent e) {
						abrirFichero(new File(elem), null);
					}
				});

				menuFichero.add(elemHistorial);
			}
		}

		menuFichero.addSeparator();
		menuFichero.add(new JMenuItem(accionSalir));
	}

	/**
	 * Añade los elementos necesarios al menú Fichero. Útil para actualizar sus
	 * contenidos al cambiar el modelo.
	 * 
	 * @param menuPrefs
	 *            Menú al que modificar sus elementos.
	 */
	private void sincronizarMenuPreferencias(JMenu menuPrefs) {
		menuPrefs.removeAll();

		menuHojaPreprocesado = new RadioSelectMenu(accionSelHojaPreproc,
				modPresentacion.getNombresHojasPreproc(), 
				getDocumentoActual().getNombreHojaPreproc());
		menuHojaVisualizacion = new RadioSelectMenu(accionSelHojaVista,
				modPresentacion.getNombresHojasVista(), 
				getDocumentoActual().getNombreHojaVista());
		menuPrefs.add(menuHojaPreprocesado);
		menuPrefs.add(menuHojaVisualizacion);
		menuPrefs.addSeparator();

		chkHabilitarAutoImportar = new JCheckBoxMenuItem(
				accionActivarAutoImport);
		sincronizarAutoImportar();

		menuPrefs.add(chkHabilitarAutoImportar);
		menuPrefs.addSeparator();

		menuPrefs.add(new JMenuItem(accionEditarFormatos));

		// AÑADIR DIÁLOGO FILTROS
	}

	/* -- SINCRONIZACION: MEDIO NIVEL -- */

	private void sincronizarEstado() {
		if (modPresentacion.getOcupado()) {
			progresoAccion.setValue(0);
			progresoAccion.setIndeterminate(true);
		} else {
			progresoAccion.setValue(100);
			progresoAccion.setIndeterminate(false);
		}

		estadoPrograma.setText(modPresentacion.getMensajeEstado());
	}

	private void sincronizarDocumento() {
		final DocumentPresentationModel modDocActual = getDocumentoActual();

		// Marcamos las hojas usadas en ambos menús
		menuHojaPreprocesado.setSeleccion(modDocActual.getNombreHojaPreproc());
		menuHojaVisualizacion.setSeleccion(modDocActual.getNombreHojaVista());

		setTitle(modPresentacion.getTitulo(getDocumentoActual()));

		sincronizarAutoImportar();
		habilitarManipulacion(modDocActual.getNavegable());
		if (listaDocs.getSelectedTab() != null) {
			((DocumentTab) listaDocs.getSelectedTab()).sincronizar(modDocActual);
		}
	}

	/* -- SINCRONIZACION: BAJO NIVEL -- */

	/**
	 * Sincroniza sólo la marca de autoimportación.
	 */
	private void sincronizarAutoImportar() {
		chkHabilitarAutoImportar.setSelected(modPresentacion
				.getActualizacionAutomatica());
	}

	/**
	 * Activa o desactiva todas las acciones que permiten operar sobre el
	 * documento actual.
	 * 
	 * @param b
	 *            Indica si se activan (<code>true</code>) o desactivan (<code>false</code>)
	 *            los controles.
	 */
	private void habilitarManipulacion(boolean b) {
		accionPrimerHijo.setEnabled(b);
		accionContraerTodo.setEnabled(b);
		accionEditar.setEnabled(b);
		accionExpandirTodo.setEnabled(b);
		accionBuscar.setEnabled(b);
		accionBuscarSig.setEnabled(b);
		accionHermanoSig.setEnabled(b);
		accionPadre.setEnabled(b);
		accionHermanoAnt.setEnabled(b);
		accionSelHojaPreproc.setEnabled(b);
		accionSelHojaVista.setEnabled(b);
	}

	/** --- OTRAS OPERACIONES --- */

	/**
	 * Abre el documento con fichero <code>fuente</code>, posiblemente
	 * haciendo una búsqueda por la ruta <code>xpointer</code> tras ello.
	 * 
	 * @param fuente
	 *            Fichero a abrir.
	 * @param xpath
	 *            Ruta XPath a seleccionar (opcional).
	 */
	private void abrirFichero(final File fuente, final String xpath) {
		servEjecucion.ejecutar(new Callable<DocumentPresentationModel>() {
			public DocumentPresentationModel call() throws Exception {
				return modPresentacion.solicitudAbrirDocumento(fuente);
			}
		}, new UnaryFunctor<DocumentPresentationModel>() {
			public void ejecutar(final DocumentPresentationModel nDoc) {
				try {
					final boolean nuevoDocumento = modPresentacion
							.agregarDocumento(nDoc);
					nDoc.getNotCambios().agregarObservadorCambioFichero(
							WndMain.this);

					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							try {
								sincronizarMenuFichero(menuFichero);
								if (nuevoDocumento) {
									listaDocs.agregarTab(fuente);
									listaDocs.seleccionarTab(fuente);

									final DocumentView vista = listaDocs
											.getVistaTabActual();
									vista
											.addTreeSelectionListener(WndMain.this);
									vista.addHyperlinkListener(WndMain.this);
								} else {
									listaDocs.seleccionarTab(fuente);
								}
								sincronizarDocumento();

								if (xpath != null) {
									SearchRequest ev = new SearchRequest();
									ev.setRutaXPath(xpath);
									busquedaLanzada(ev);
								}
							} catch (Exception ex) {
								ex.printStackTrace();
							}
						}
					});
				} catch (TransformationException ex) {
					ex.printStackTrace();
				}
			}
		}, new UnaryFunctor<Exception>() {
			public void ejecutar(Exception ex) {
				ex.printStackTrace();
				MessageManagerSingleton.getInstancia().mostrarError(
						errRes.getString("err_open_fail") + "\n"
								+ ex.getLocalizedMessage());
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						sincronizarMenuFichero(menuFichero);
					}
				});
			}
		});

	}

	/* --- CLASES PRIVADAS ------------------------------ */

	/* ACCIONES - sólo las más complejas */

	/** Acción de apertura de documentos. */
	private class OpenAction extends ResourceBackedAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = -3033019833259042023L;

		public OpenAction() {
			super(res, "action_open", res.getString("action_open_i"));
		}

		public void actionPerformed(ActionEvent e) {
			// Creamos un selector de ficheros actualizado con los
			// formatos admitidos actualmente
			JFileChooser dlgAbrir = new JFileChooser();
			dlgAbrir.setMultiSelectionEnabled(false);

			final List<Pair<String, String[]>> formatos = modPresentacion
					.getElemsSelectorFicheros();
			for (Pair<String, String[]> entrada : formatos) {
				dlgAbrir.addChoosableFileFilter(new ExtensionFileFilter(entrada
						.segundo(), entrada.primero()));
			}

			if (dlgAbrir.showOpenDialog(WndMain.this) == JFileChooser.APPROVE_OPTION) {
				abrirFichero(dlgAbrir.getSelectedFile(), null);
			}
		}
	}

	/** Acción de lanzamiento de diálogo Buscar. */
	private class FindAction extends ResourceBackedAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = -5295820489091250276L;

		public FindAction() {
			super(res, "action_find", res.getString("action_find_i"));
		}

		public void actionPerformed(ActionEvent e) {
			if (vntBusqueda == null) {
				vntBusqueda = new WndFind(WndMain.this);
				vntBusqueda.agregarObservadorBusqueda(WndMain.this);
			}

			vntBusqueda.setVisible(true);
		}
	}

	/** Acción de búsqueda de la siguiente coincidencia. */
	private class FindNextAction extends ResourceBackedAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = 6994490721852391963L;

		public FindNextAction() {
			super(res, "action_findnext");
		}

		public void actionPerformed(ActionEvent e) {
			final DocumentPresentationModel docActual = getDocumentoActual();

			servEjecucion.ejecutar(new Callable<NodeList>() {
				public NodeList call() throws Exception {
					return docActual.repetirBusqueda();
				}
			}, new UnaryFunctor<NodeList>() {
				public void ejecutar(final NodeList res) {
					docActual.finBusqueda(res, null);
					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							if (res == null) {
								MessageManagerSingleton
										.getInstancia()
										.mostrarError(
												errRes
														.getString("err_findnext_nofindyet"));
							} else if (res.getLength() == 0) {
								MessageManagerSingleton
										.getInstancia()
										.mostrarError(
												miscRes
														.getString("find_nomatches"));
							}
							sincronizarDocumento();
						}
					});
				}
			}, new UnaryFunctor<Exception>() {
				public void ejecutar(Exception ex) {
					MessageManagerSingleton.getInstancia().mostrarError(
							ex.getMessage());
				}
			});

		}
	}

	/** Acción de salida del programa. */
	private class ExitAction extends ResourceBackedAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = -5183025387352671187L;

		public ExitAction() {
			super(res, "action_exit", res.getString("action_exit_i"));
		}

		public void actionPerformed(ActionEvent e) {
			if (MessageManagerSingleton.getInstancia().preguntarSiNo(
					res.getString("confirm_exit_msg"))) {
				aplicacion.close();
			}
		}
	}

	/** Acción de cambio de hoja de preprocesado */
	private class ChangePreprocessingStylesheetAction extends ResourceBackedAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = -5434085806022651386L;

		public ChangePreprocessingStylesheetAction() {
			super(res, "action_selectstylesheet_preproc");
		}

		public void actionPerformed(final ActionEvent e) {
			try {
				modPresentacion.cambioHojaPreprocesado(getDocumentoActual(), e.getActionCommand());
				sincronizarDocumento();
			} catch (TransformationException ex) {
				MessageManagerSingleton.getInstancia().mostrarError(
						ex.getMessage());
				sincronizarDocumento();
			} catch (ParsingException ex) {
				MessageManagerSingleton.getInstancia().mostrarError(
						ex.getMessage());
				sincronizarDocumento();
			}
		}
	}

	/** Acción de cambio de hoja de visualización */
	private class ChangeViewStylesheetAction extends ResourceBackedAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = 361078597085383631L;

		public ChangeViewStylesheetAction() {
			super(res, "action_selectstylesheet_view");
		}

		public void actionPerformed(ActionEvent e) {
			try {
				modPresentacion.cambioHojaVisualizacion(
						getDocumentoActual(),
						e.getActionCommand());
				sincronizarDocumento();
			} catch (TransformationException ex) {
				MessageManagerSingleton.getInstancia().mostrarError(
						ex.getMessage());
				sincronizarDocumento();
			}
		}
	}

	/** Acción de edición de formatos disponibles para el usuario */
	private class EditFormatsAction extends ResourceBackedAction {
		/**
		 * 
		 */
		private static final long serialVersionUID = -8226394759881245434L;

		public EditFormatsAction() {
			super(res, "action_editformats");
		}

		public void actionPerformed(ActionEvent e) {
			new WndFormats(WndMain.this, modPresentacionFormatos);
		}
	}

	@Override
	public void tabAdded(TabEvent arg0) {}

	@Override
	public void tabDehighlighted(TabStateChangedEvent arg0) {}

	@Override
	public void tabDeselected(TabStateChangedEvent arg0) {}

	@Override
	public void tabDragAborted(TabEvent arg0) {}

	@Override
	public void tabDragged(TabDragEvent arg0) {}

	@Override
	public void tabDropped(TabDragEvent arg0) {}

	@Override
	public void tabHighlighted(TabStateChangedEvent arg0) {}

	@Override
	public void tabMoved(TabEvent arg0) {}

	@Override
	public void tabRemoved(TabRemovedEvent arg0) {
		 modPresentacion.cierreDocumento(((DocumentTab)arg0.getTab()).getFichero());
	}

	@Override
	public void tabSelected(TabStateChangedEvent arg0) {
		sincronizarDocumento();
	}

	private DocumentPresentationModel getDocumentoActual() {
		DocumentTab tabActual = (DocumentTab)listaDocs.getSelectedTab();
		File fichActual       = tabActual != null ? tabActual.getFichero() : null;
		return modPresentacion.getModeloDocumento(fichActual);
	}
}
