package es.uca.xmleye.gui.util;

/*
 Copyright (C) 2005-2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.io.File;
import javax.swing.filechooser.FileFilter;

/**
 * Filtro de ficheros basado en una o más extensiones. Acepta todas las
 * carpetas.
 */
public class ExtensionFileFilter extends FileFilter {
	private String[] filterExts;
	private String desc;

	/**
	 * Constructor para una sola extensión.
	 * 
	 * Método de conveniencia para
	 * <code>ExtensionFileFilter(String[],String)</code>.
	 * 
	 * @param ext
	 *            Extensión a aceptar.
	 * @param desc
	 *            Descripción del filtro.
	 */
	public ExtensionFileFilter(String ext, String desc) {
		cargar(new String[] { ext }, desc);
	}

	/**
	 * Constructor para una sola extensión.
	 * 
	 * @param exts
	 *            Extensiones a aceptar.
	 * @param desc
	 *            Descripción del filtro.
	 */
	public ExtensionFileFilter(String[] exts, String desc) {
		cargar(exts, desc);
	}

	@Override
	public boolean accept(File f) {
		if (f.isDirectory())
			return true;
		String fileExt = getExtension(f.getName());

		for (String ext : filterExts) {
			if (fileExt.equals(ext))
				return true;
		}
		return false;
	}

	@Override
	public String getDescription() {
		return desc;
	}

	public String[] getExtensionesFiltro() {
		return filterExts;
	}

	private void cargar(String[] exts, String desc) {
		this.filterExts = exts;
		this.desc = desc;
	}

	private String getExtension(String file) {
		int ultimoPunto = file.lastIndexOf('.');
		if (ultimoPunto == -1)
			return "";
		else {
			return file.substring(ultimoPunto + 1).toLowerCase();
		}
	}
}
