package es.uca.xmleye.util;

/*
 Copyright (C) 2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.util.concurrent.Callable;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;

/**
 * Integramos un sistema de ejecución de tareas en segundo plano con la
 * capacidad de ejecutar callback en caso de error y en caso de éxito.
 */
public class MultipleCallbackExecutorService {
	// Colección de hilos reutilizables
	private ExecutorService ejecutor;

	public MultipleCallbackExecutorService() {
		this.ejecutor = Executors.newSingleThreadExecutor();
	}

	// Empleamos el ejecutor anidado, y añadimos algo de código por nuestra
	// cuenta
	public synchronized <V> void ejecutar(final Callable<V> orden,
			final UnaryFunctor<V> callbackExito,
			final UnaryFunctor<Exception> callbackFallo) {
		ejecutor.submit(new Runnable() {
			public void run() {
				try {
					V resultado = orden.call();
					callbackExito.ejecutar(resultado);
				} catch (InterruptedException ex) {
				} catch (Exception ex) {
					callbackFallo.ejecutar(ex);
				}
			}
		});
	}

}
